from typing import List, Union, Tuple
import cadquery as cq
from OCP.BRepBuilderAPI import BRepBuilderAPI_MakeEdge

from .geometry import Pnt, Dir, Lin
from .api import check_overlap, intersect
from ..types import Real, Vector3
from .util import get_components, get_object
from cadquery.occ_impl.geom import BoundBox


class Track(object):
    def __init__(self, origin: Vector3, dir: Vector3, local: cq.Plane = cq.Plane((0,0,0), normal = (0,0,1))):
        self.origin: Pnt = Pnt(*origin)
        self.dir: Dir = Dir(*dir)
        self.line = Lin(self.origin, self.dir)
        self.local = local


    def local_xy(self, z: Real) -> cq.Vector:
        return cq.Vector(self.origin) + (z - self.origin.Z()) * cq.Vector(self.dir)


    def make_edge_local(self, l: Real) -> cq.Edge:
        return cq.Edge(BRepBuilderAPI_MakeEdge(self.line, 0, l).Edge())


    def make_edge(self, l: Real) -> cq.Shape:
        return self.make_edge_local(l).transformShape(self.local.rG)


class TrackIntersection(object):
    def __init__(
            self,
            edge: cq.Shape,
            track: Track,
            object: cq.Assembly
        ) -> None:
        self.edge: cq.Shape = edge
        self.track: Track = track
        self.object: cq.Assembly = object
        self.dist: Real = self.edge.Center().toPnt().Distance(
            self.track.origin
        )
        
    
    def __lt__(self, other):
        return self.dist < other.dist
    
    
    def __gt__(self, other):
        return self.dist > other.dist


def intersect_assembly_track(
        assembly: cq.Assembly,
        track: Track,
        l: Real) ->  List[TrackIntersection]:
    results: List[TrackIntersection] = []

    edge = track.make_edge(l)
    bb_edge: BoundBox = edge.BoundingBox()

    for x in get_components(assembly).values():
        obj = get_object(x, coord='world').wrapped
        if check_overlap(bb_edge, obj):
            intersection_shape = intersect(
                obj,
                edge.wrapped
            )
            for e in intersection_shape.Edges():
                results.append(TrackIntersection(
                    e,
                    track,
                    x
                ))

    results.sort()
    return results



from abc import abstractmethod
from typing import Dict, List, Optional, Any
import cadquery as cq
import logging

from ..types import Real, Vector3
from ..core.track import Track, TrackIntersection, intersect_assembly_track
from ..core.util import get_components
from ..core.material import MaterialLike

logger = logging.getLogger(__name__)


def make_metadata(
    material: Optional[MaterialLike] = None
):
    metadata = {}
    if material:
        metadata['material'] = material
    return metadata


class SetupBase(object):
    def __init__(
        self,
        name: Optional[str] = None,
        origin: Vector3 = (0.0, 0.0, 0.0),
        normal: Vector3 = (0.0, 0.0, 1.0),
        xDir: Vector3 = (1.0, 0.0, 0.0),
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        self.setup = cq.Assembly(
            name = name,
            metadata = metadata)
        self.construct(
            origin = origin,
            normal = normal,
            xDir = xDir,
            **kwargs
        )

    @abstractmethod
    def construct(
        self,
        origin: Vector3 = (0.0, 0.0, 0.0),
        normal: Vector3 = (0.0, 0.0, 1.0),
        xDir: Vector3 = (1.0, 0.0, 0.0),
        **kwargs
    ):
        '''
        example

        self.setup.add(
            cq.Workplane().box(100, 100, 100),
            name='box',
            metadata=make_metadata(material=gagg)
        )
        '''
        pass


    def add(self, *args, **kwargs):
        self.setup.add(*args, **kwargs)


    def intersect_track(self, track: Track, l: Real = 10e3) ->  List[TrackIntersection]:
        return intersect_assembly_track(self.setup, track, l)
    

    def components(self) -> Dict[str, cq.Assembly]:
        return get_components(self.setup)






from . import setup
from . import physics
from . import core
from .core import Track, pol2xyz, xyz2pol, pol, xyz, Material, MaterialLayers
from . import setup
from .setup import SetupBase, make_metadata



from typing import Union
import cadquery as cq
from OCP.TopoDS import TopoDS_Shape
from OCP.BRepAlgoAPI import BRepAlgoAPI_Common
from ..error import OCPOperationFailError
from cadquery.occ_impl.geom import BoundBox


def intersect(obj1: TopoDS_Shape, obj2: TopoDS_Shape) -> cq.Shape:
    com = BRepAlgoAPI_Common(obj1, obj2)
    com.Build()
    if not com.IsDone():
        raise OCPOperationFailError('failed to calculate the intersection')
    return cq.Shape(com.Shape())


def subtract():
    ...


def union():
    ...


def check_overlap(
          obj1: Union[TopoDS_Shape, cq.Shape, BoundBox],
          obj2: Union[TopoDS_Shape, cq.Shape, BoundBox]) -> bool:
    if isinstance(obj1, TopoDS_Shape):
        bb_obj1: BoundBox = cq.Shape(obj1).BoundingBox()
    elif isinstance(obj1, cq.Shape):
        bb_obj1: BoundBox = obj1.BoundingBox()
    else:
        bb_obj1: BoundBox = obj1
    
    if isinstance(obj2, TopoDS_Shape):
        bb_obj2: BoundBox = cq.Shape(obj2).BoundingBox()
    elif isinstance(obj2, cq.Shape):
        bb_obj2: BoundBox = obj2.BoundingBox()
    else:
        bb_obj2: BoundBox = obj2

    overlap_x = not (bb_obj1.xmax < bb_obj2.xmin or bb_obj1.xmin > bb_obj2.xmax)
    overlap_y = not (bb_obj1.ymax < bb_obj2.ymin or bb_obj1.ymin > bb_obj2.ymax)
    overlap_z = not (bb_obj1.zmax < bb_obj2.zmin or bb_obj1.zmin > bb_obj2.zmax)
    return overlap_x and overlap_y and overlap_z



from .geometry import (
    XYZ,
    Pnt,
    Vec,
    Dir,
    Lin,
    Ax1,
    Ax2,
    pol2xyz,
    xyz2pol,
    pol,
    xyz
)
from .material import Material, MaterialLayers
from .track import Track


from typing import Dict, Literal, Union
import cadquery as cq


def get_components(assembly: cq.Assembly) -> Dict[str, cq.Assembly]:
        results = {}
        for x in assembly.children:
            if x.children:
                results.update(get_components(x))
            else:
                results[x.name] = x
        return results


def get_object(assy: cq.Assembly, coord: Literal['world', 'local'] = 'world') -> cq.Shape:
    obj = assy.obj

    if isinstance(obj, cq.Workplane):
        obj = obj.val()
    elif isinstance(obj, cq.Shape):
        pass
    else:
        raise RuntimeError(f'{assy.name} has no shape')

    if coord == 'world':
        loc = assy.loc * obj.location() # type: ignore
        while assy.parent is not None:
            assy = assy.parent
            loc = assy.loc * loc
        obj = obj.located(loc) # type: ignore

    if isinstance(obj, cq.Shape):
        return obj
    else:
        raise RuntimeError(f'returned object is not Shape')



import numpy as np

class kinema():
    def __init__(self, m1, m2, m3, m4, num=1000):
        self.m1 = m1
        self.m2 = m2
        self.m3 = m3
        self.m4 = m4
        self.theta_cm = np.linspace(1e-5, np.pi-1e-5, num)
        self.p_cm = 0.0
        self.T3_cm = 0.0
        self.T4_cm = 0.0
        self.theta3_lab = np.zeros(num)
        self.theta4_lab = np.zeros(num)
        self.T3_lab = np.zeros(num)
        self.T4_lab = np.zeros(num)
        self.p3_lab = np.zeros(num)
        self.p4_lab = np.zeros(num)
        self.beta3_lab = np.zeros(num)
        self.beta4_lab = np.zeros(num)

    def calc(self, T):
        m1 = self.m1
        m2 = self.m2
        m3 = self.m3
        m4 = self.m4
        theta_cm = self.theta_cm
        
        sq = lambda x : np.power(x,2)

        # mandelstam variable s
        s = sq(m1 + m2) + 2*m1*T
        # Momentum in the CM
        pcm12 = np.sqrt( (sq(s - sq(m1) - sq(m2)) - 4*sq(m1)*sq(m2)) / (4*s) )
        gamma = np.sqrt(sq(m1) + sq(pcm12)) / m1 # coshX
        beta_gamma = pcm12 / m1 # sinhX

        # Energy of particles 3 & 4 in CM
        E3_cm = ( s + (sq(m3) - sq(m4)) ) / (2 * np.sqrt(s))
        E4_cm = ( s - (sq(m3) - sq(m4)) ) / (2 * np.sqrt(s))
        
        # Kinetic energy of particles 3 & 4 in CM
        self.T3_cm = E3_cm - m3
        self.T4_cm = E4_cm - m4
        
        # Momentum in the CM
        self.p_cm = np.sqrt( (sq(s - sq(m3) - sq(m4)) - 4*sq(m3)*sq(m4)) / (4*s) )
        p_cm = self.p_cm
        
        # Energy of particles 3 & 4 in Lab
        E3_lab = gamma * E3_cm + beta_gamma * p_cm * np.cos(theta_cm)
        E4_lab = gamma * E4_cm - beta_gamma * p_cm * np.cos(theta_cm)
        
        # Kinetic energy of particles 3 & 4 in Lab
        self.T3_lab = E3_lab - m3
        self.T4_lab = E4_lab - m4
        
        # Momentum in Lab
        p3cos = gamma * p_cm * np.cos(theta_cm) + E3_cm * beta_gamma
        p4cos = - gamma * p_cm * np.cos(theta_cm) + E4_cm * beta_gamma
        p3sin = p_cm * np.sin(theta_cm)
        p4sin = p_cm * np.sin(theta_cm)
        self.p3_lab = np.sqrt(sq(p3cos) + sq(p3sin))
        self.p4_lab = np.sqrt(sq(p4cos) + sq(p4sin))

        # Beta in Lab
        gamma3_lab = E3_lab/m3
        gamma4_lab = E4_lab/m4
        self.beta3_lab = np.sqrt( (sq(gamma3_lab) - 1)/sq(gamma3_lab) )
        self.beta4_lab = np.sqrt( (sq(gamma4_lab) - 1)/sq(gamma4_lab) )

        # Theta in Lab
        self.theta3_lab = np.arctan(p3sin/p3cos)
        self.theta4_lab = np.arctan(p4sin/p4cos)


from typing import NewType, List, Optional, Union, overload
import pycatima as catima
import logging

from ..types import Real
from .track import TrackIntersection


logger = logging.getLogger(__name__)


class Material(catima.Material): # type: ignore
    def __init__(self, *args, name: str, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.name: str = name
        self.components = []
        for i in range(self.ncomponents()):
            comp = self.get_element(i)
            self.components.append([
                comp.A,
                comp.Z,
                comp.stn
            ])

    def __reduce__(self):
        return (
            Material._reconstruct,
            (
                self.name,
                self.components,
                self.density(),
                self.I(),
                self.thickness()
            ),
            None
        )

    @staticmethod
    def _reconstruct(
        name: str,
        components: List,
        density: Real,
        i_potential: Real,
        thickness: Real):
        return Material(
            components,
            name=name,
            density=density,
            i_potential=i_potential,
            thickness=thickness)


# Define types for material
CATIMA_Material = NewType('CATIMA_Material', catima.Material)  # type: ignore
NC_Material = NewType('NC_Material', Material)
MaterialLike = Union[CATIMA_Material, NC_Material]


class MaterialLayers(catima.Layers): # type: ignore

    @overload
    def __init__(
        self,
        arg: Optional[List[TrackIntersection]] = None) -> None:
        ...


    @overload
    def __init__(
        self,
        arg: Optional[List[Material]] = None) -> None:
        ...


    def __init__(
            self,
            arg = None) -> None:
        super().__init__()
        self.names: List[str] = []
        if arg:
            if all(isinstance(x, TrackIntersection) for x in arg):
                self.add_intersections(arg)
            elif all(isinstance(x, Material) for x in arg):
                self.add_materials(arg)
            else:
                raise RuntimeError('arg should be list of TrackIntersection or Material')


    def add(self, material: MaterialLike, name: Optional[str] = None) -> None:
        if type(material) is Material:
            super().add(material)
            if name:
                self.names.append(name)
            else:
                self.names.append(material.name)
        else:
            super().add(material)
            if name:
                self.names.append(name)
            else:
                raise RuntimeError('name is required.')


    def add_intersections(self, intersections: List[TrackIntersection]) -> None:
        for x in intersections:
            name = x.object.name
            if 'material' in x.object.metadata:
                mat = x.object.metadata['material']
                thickness = x.edge.Length() # type: ignore
                mat.thickness_cm(thickness * 1e-1) # mm->cm
                self.add(mat, name=name)
            else:
                logger.info(f'ignore {x.object.name} because the object does not contain material in metadata.')


    def add_materials(self, materials: List[Material]) -> None:
        for x in materials:
            self.add(x)


    def size(self) -> int:
        return self.num()
    

    def __len__(self) -> int:
        return self.size()


    def __sizeof__(self) -> int:
        return self.size()


    @overload
    def get(self, key: str) ->  Material:
        ...


    @overload
    def get(self, key: int) -> Material:
        ...


    def get(self, key) -> Material:
        if isinstance(key, int):
            idx = key
        elif isinstance(key, str):
            if key in self.names:
                idx = self.names.index(key)
            else:
                raise KeyError(f'{key} not found')
        else:
            raise TypeError
        return Material(super().get(idx), name=self.names[idx])


    def to_list(self) -> List[Material]:
        mat_list = []
        for i in range(self.size()):
            mat_list.append(self.get(i))
        return mat_list


    def __reduce__(self):
        return (
            MaterialLayers._reconstruct,
            tuple([self.to_list()]),
            None
        )


    @staticmethod
    def _reconstruct(materials: List[Material]):
        return MaterialLayers(materials)


# Define types for layers of Materials
CATIMA_Layers = NewType('CATIMA_Layers', catima.Layers) # type: ignore
NC_Layers = NewType('NC_Layers', MaterialLayers)
LayersLike = Union[CATIMA_Layers, NC_Layers]


import math
from OCP.gp import gp_XYZ, gp_Pnt, gp_Vec, gp_Dir, gp_Lin, gp_Ax1, gp_Ax2
from ..types import Real, Vector3


def pol2xyz(r: Real, theta: Real, phi: Real) -> Vector3:
    x = r * math.sin(theta) * math.cos(phi)
    y = r * math.sin(theta) * math.sin(phi)
    z = r * math.cos(theta)
    return (x, y, z)


def xyz2pol(x: Real, y: Real, z: Real) -> Vector3:
    r = math.sqrt(x**2 + y**2 + z**2)
    theta = 0.0
    phi = 0.0
    if r != 0:
        theta = math.acos(z / r)
        phi = math.copysign(1,y) * math.acos(x/math.sqrt(x**2+y**2))
    return (r, theta, phi)


def xyz(x: Real, y: Real, z: Real) -> Vector3:
    return (x,y,z)


def pol(r: Real, theta: Real, phi: Real) -> Vector3:
    return pol2xyz(r, theta, phi)


class XYZ(gp_XYZ):
    '''
    Wrapper of gp_XYZ
    '''
    pass


class Pnt(gp_Pnt):
    '''
    Wrapper of gp_Pnt
    '''
    pass


class Vec(gp_Vec):
    '''
    Wrapper of gp_Vec
    '''
    pass

class Dir(gp_Dir):
    '''
    Wrapper of gp_Dir
    '''
    pass


class Ax1(gp_Ax1):
    '''
    Wrapper of gp_Ax1
    '''
    pass


class Ax2(gp_Ax2):
    '''
    Wrapper of gp_Ax2
    '''
    pass


class Lin(gp_Lin):
    '''
    Wrapper of gp_Lin
    '''
    pass


